@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 module Test.Hspec.Core.HooksSpec (spec) where
 
@@ -17,8 +18,9 @@ import qualified Test.Hspec.Core.Hooks as H
 evalSpec_ :: H.Spec -> IO ()
 evalSpec_ = void . evalSpec
 
+
 evalSpec :: H.Spec -> IO [([String], Item)]
-evalSpec = fmap normalize . (fmap (H.specToEvalForest H.defaultConfig) . H.runSpecM >=> runFormatter config)
+evalSpec = fmap normalize . (toEvalForest >=> runFormatter config)
   where
     config = EvalConfig {
       evalConfigFormat = \ _ -> return ()
@@ -36,6 +38,9 @@ evalSpec = fmap normalize . (fmap (H.specToEvalForest H.defaultConfig) . H.runSp
     }
     pathToList (xs, x) = xs ++ [x]
 
+toEvalForest :: H.SpecWith () -> IO [EvalTree]
+toEvalForest = fmap (uncurry H.specToEvalForest . first (($ H.defaultConfig) . appEndo)) . H.runSpecM
+
 mkAppend :: IO (String -> IO (), IO [String])
 mkAppend = do
   ref <- newIORef ([] :: [String])
@@ -142,7 +147,7 @@ spec = do
             n `shouldBe` 23
         `shouldReturn` [
           item ["foo"] divideByZero
-        , item ["bar"] (Pending Nothing (Just "exception in beforeAll-hook (see previous failure)"))
+        , item ["bar"] (Pending Nothing $ exceptionIn "beforeAll")
         ]
 
     context "when used with an empty list of examples" $ do
@@ -233,7 +238,7 @@ spec = do
                 n `shouldBe` 23
         `shouldReturn` [
           item ["foo"] divideByZero
-        , item ["bar"] (Pending Nothing (Just "exception in beforeAll-hook (see previous failure)"))
+        , item ["bar"] (Pending Nothing $ exceptionIn "beforeAllWith")
         ]
 
     context "when used with an empty list of examples" $ do
@@ -317,7 +322,6 @@ spec = do
         , "foo"
         , "before"
         , "bar"
-        , "before"
         , "from before"
         ]
 
@@ -331,9 +335,10 @@ spec = do
       it "reports a failure" $ do
         evalSpec $ H.before (return "from before") $ H.afterAll (\_ -> throwException) $ do
           H.it "foo" $ \a -> a `shouldBe` "from before"
+          H.it "bar" $ \a -> a `shouldBe` "from before"
         `shouldReturn` [
           item ["foo"] Success
-        , item ["afterAll-hook"] divideByZero
+        , item ["bar"] $ divideByZeroIn "afterAll"
         ]
 
   describe "afterAll_" $ do
@@ -349,7 +354,6 @@ spec = do
         , "foo"
         , "before"
         , "bar"
-        , "before"
         , "afterAll_"
         ]
 
@@ -377,9 +381,10 @@ spec = do
         evalSpec $ do
           H.afterAll_ H.pending $ do
             H.it "foo" True
+            H.it "bar" True
         `shouldReturn` [
           item ["foo"] Success
-        , item ["afterAll-hook"] (Pending Nothing Nothing)
+        , item ["bar"] (Pending Nothing Nothing)
         ]
 
     context "when action throws an exception" $ do
@@ -387,13 +392,14 @@ spec = do
         evalSpec $ do
           H.afterAll_ throwException $ do
             H.it "foo" True
+            H.it "bar" True
         `shouldReturn` [
           item ["foo"] Success
-        , item ["afterAll-hook"] divideByZero
+        , item ["bar"] $ divideByZeroIn "afterAll_"
         ]
 
     context "when action is successful" $ do
-      it "does not report anythig" $ do
+      it "does not report anything" $ do
         evalSpec $ do
           H.afterAll_ (return ()) $ do
             H.it "foo" True
@@ -497,6 +503,36 @@ spec = do
         , "after"
         ]
 
+    it "wrap actions around a spec in order" $ do
+      (rec, retrieve) <- mkAppend
+      let action i inner = rec ("before " <> i) *> inner <* rec ("after " <> i)
+      evalSpec_ $ H.aroundAll_ (action "1") $ H.aroundAll_ (action "2") $ do
+        H.it "foo" $ rec "foo"
+        H.it "bar" $ rec "bar"
+      retrieve `shouldReturn` [
+          "before 1"
+        , "before 2"
+        , "foo"
+        , "bar"
+        , "after 2"
+        , "after 1"
+        ]
+
+    it "does not call actions wrapped around a failing action" $ do
+      (rec, retrieve) <- mkAppend
+      let action i inner = rec ("before " <> i) *> inner <* rec ("after " <> i)
+      evalSpec_ $
+        H.aroundAll_ (action "1") $
+          H.aroundAll_ (action "2 failing" . const throwException) $
+            H.aroundAll_ (action "3") $ do
+              H.it "foo" $ rec "foo"
+              H.it "bar" $ rec "bar"
+      retrieve `shouldReturn` [
+          "before 1"
+        , "before 2 failing"
+        , "after 1"
+        ]
+
     it "does not memoize subject" $ do
       mock <- newMock
       let action :: IO Int
@@ -512,24 +548,26 @@ spec = do
         , "2"
         , "3"
         ]
-      mockCounter mock `shouldReturn` 4
+      mockCounter mock `shouldReturn` 3
 
     it "reports exceptions on acquire" $ do
       evalSpec $ do
         H.aroundAll_ (throwException <*) $ do
           H.it "foo" True
+          H.it "bar" True
       `shouldReturn` [
         item ["foo"] divideByZero
-      , item ["afterAll-hook"] (Pending Nothing (Just "exception in beforeAll-hook (see previous failure)"))
+      , item ["bar"] (Pending Nothing $ exceptionIn "aroundAll_")
       ]
 
     it "reports exceptions on release" $ do
       evalSpec $ do
         H.aroundAll_ (<* throwException) $ do
           H.it "foo" True
+          H.it "bar" True
       `shouldReturn` [
         item ["foo"] Success
-      , item ["afterAll-hook"] divideByZero
+      , item ["bar"] $ divideByZeroIn "aroundAll_"
       ]
 
   describe "aroundAllWith" $ do
@@ -546,7 +584,37 @@ spec = do
         , "1"
         , "1"
         ]
-      mockCounter mock `shouldReturn` 4
+      mockCounter mock `shouldReturn` 3
+
+    it "wrap actions around a spec in order" $ do
+      (rec, retrieve) <- mkAppend
+      let action i inner a = rec ("before " <> i) *> inner a <* rec ("after " <> i)
+      evalSpec_ $ H.aroundAllWith (action "1") $ H.aroundAllWith (action "2") $ do
+        H.it "foo" $ rec "foo"
+        H.it "bar" $ rec "bar"
+      retrieve `shouldReturn` [
+          "before 1"
+        , "before 2"
+        , "foo"
+        , "bar"
+        , "after 2"
+        , "after 1"
+        ]
+
+    it "does not call actions wrapped around a failing action" $ do
+      (rec, retrieve) <- mkAppend
+      let action i inner a = rec ("before " <> i) *> inner a <* rec ("after " <> i)
+      evalSpec_ $
+        H.aroundAllWith (action "1") $
+          H.aroundAllWith (action "2 failing" . const . const throwException) $
+            H.aroundAllWith (action "3") $ do
+              H.it "foo" $ rec "foo"
+              H.it "bar" $ rec "bar"
+      retrieve `shouldReturn` [
+          "before 1"
+        , "before 2 failing"
+        , "after 1"
+        ]
 
     it "reports exceptions on acquire" $ do
       evalSpec $ do
@@ -554,21 +622,63 @@ spec = do
           H.it "foo" H.pending
       `shouldReturn` [
         item ["foo"] divideByZero
-      , item ["afterAll-hook"] (Pending Nothing (Just "exception in beforeAll-hook (see previous failure)"))
       ]
 
     it "reports exceptions on release" $ do
       evalSpec $ do
         H.aroundAllWith (\ action () -> action () <* throwException) $ do
           H.it "foo" True
+          H.it "bar" True
       `shouldReturn` [
         item ["foo"] Success
-      , item ["afterAll-hook"] divideByZero
+      , item ["bar"] $ divideByZeroIn "aroundAllWith"
       ]
 
+  describe "decompose" $ do
+    it "decomposes a with-style action into acquire / release" $ do
+      (acquire, release) <- H.decompose $ \ action x -> do
+        action (x + 42 :: Int)
+      acquire 23 `shouldReturn` 65
+      release
+
+    context "when release is called before acquire" $ do
+      it "does nothing" $ do
+        (_, release) <- H.decompose $ \ action x -> do
+          action (x + 42 :: Int)
+        release
+
+    context "with an exception during resource acquisition" $ do
+      it "propagates that exception" $ do
+        (acquire, release) <- H.decompose $ \ action () -> do
+          throwException_
+          action ()
+        acquire () `shouldThrow` (== DivideByZero)
+        release
+
+    context "with an exception during resource deallocation" $ do
+      it "propagates that exception" $ do
+        (acquire, release) <- H.decompose $ \ action () -> do
+          action ()
+          throwException_
+        acquire ()
+        release `shouldThrow` (== DivideByZero)
+
   where
     divideByZero :: Result
     divideByZero = Failure Nothing (Error Nothing $ toException DivideByZero)
 
+    divideByZeroIn :: String -> Result
+#if MIN_VERSION_base(4,8,1)
+    divideByZeroIn hook = Failure Nothing (Error (Just $ "in " <> hook <> "-hook:") $ toException DivideByZero)
+#else
+    divideByZeroIn _ = Failure Nothing (Error Nothing $ toException DivideByZero)
+#endif
+
     item :: [String] -> Result -> ([String], Item)
     item path result = (path, Item Nothing 0 "" result)
+
+#if MIN_VERSION_base(4,8,1)
+    exceptionIn name = Just ("exception in " <> name <> "-hook (see previous failure)")
+#else
+    exceptionIn _ = Just "exception in beforeAll-hook (see previous failure)"
+#endif
